package server;

import java.util.LinkedList;
import java.util.HashSet;
import java.util.Iterator;
import java.io.*;

import server.handler.*;
import common.*;

public class Reactor {
  private LinkedList registrations;
  
  private Reactor() {
    registrations = new LinkedList();
  }
  
  public synchronized void addHandler(
    Handle handle,
    EventHandler handler,
    int event )
  {
    Registration reg = new Registration( handle, handler, event );
    System.out.println( "[Reactor] adding: " + reg );
    registrations.add( reg );
  }
  
  public synchronized EventHandler removeHandler( Handle handle, int event ) {
    Iterator it = registrations.iterator();

    while ( it.hasNext() ) {
      Registration reg = (Registration) it.next();
      
      if ( reg.isHandle( handle ) && reg.handlesEvent( event ) ) {
        System.out.println( "[Reactor] removing: " + reg );
        it.remove();
        
        if ( event != Event.ANY_EVENT )
          return reg.getHandler();
      }
    }
    
    return null;
  }
  
  /*
   * Event loop
   */
  public void handleEvents() {
    while ( true ) {
      HashSet readyHandles = new HashSet();
      LinkedList potentialRegistrations = new LinkedList();
      
      synchronized ( this ) {
        Iterator it = registrations.iterator();
      
        while ( it.hasNext() ) {
          Registration reg = (Registration) it.next();
          
          if ( reg.getHandle().ready() ) {
            readyHandles.add( reg.getHandle() );
            potentialRegistrations.add( reg );
          }
        }
      }
      
      if ( readyHandles.size() == 0 )
        // don't hammer the CPU,
        // unless there's incomming data
        Sleeper.nap(); 
      else {
        System.out.println( "[Reactor] incomming detected" );
        Iterator handleIterator = readyHandles.iterator();
        
        while ( handleIterator.hasNext() ) {
          Handle handle = (Handle) handleIterator.next();
          
          String firstLine = handle.readLine().trim();
          System.out.println( "[Server] received: " + firstLine );
          String[] tokens = firstLine.split( " " );
          
          int event = Event.nameToEvent( tokens[0] );
          
          if ( event == Event.UNKNOWN_EVENT ) {
            handle.sendError( "unknown command: " + firstLine );
            continue;
          }
          
          Iterator registrationIterator = potentialRegistrations.iterator();
          
          while ( registrationIterator.hasNext() ) {
            Registration reg = (Registration) registrationIterator.next();
            
            if ( reg.isHandle( handle ) && reg.handlesEvent( event ) ) {
              reg.getHandler().handleEvent( handle, event, tokens );
              break;
            }
          }
        }
        
        System.out.println( "[Reactor] --- finished handling event ---------" );
      }
    }
  }
  
  private class Registration {
    private Handle handle;
    private EventHandler handler;
    private int event;
    
    public Registration( Handle handle, EventHandler handler, int event ) {
      this.handle = handle;
      this.handler = handler;
      this.event = event;
    }
    
    public boolean isHandle( Handle handle ) {
      return this.handle == handle;
    }
    
    public boolean handlesEvent( int event ) {
      return this.event == event || event == Event.ANY_EVENT;
    }
    
    public boolean usesHandler( EventHandler handler ) {
      return this.handler == handler;
    }
    
    public EventHandler getHandler() {
      return handler;
    }
    
    public Handle getHandle() {
      return handle;
    }
    
    public String toString() {
      StringBuffer sb = new StringBuffer();
      
      String handlerName = handler.getClass().getName();
      if ( handlerName.indexOf( '.' ) >= 0 )
        handlerName =
          handlerName.substring( handlerName.lastIndexOf( '.' ) + 1 );
      
      sb.append( "[Registration: handler:" + handlerName );
      sb.append( ", event: " );
      
      switch ( event ) {
        case Event.GET_EVENT:
          sb.append( "GET_EVENT" );
          break;
        case Event.LOG_EVENT:
          sb.append( "LOG_EVENT" );
          break;
        case Event.LOGIN_EVENT:
          sb.append( "LOGIN_EVENT" );
          break;
        case Event.LOGOUT_EVENT:
          sb.append( "LOGOUT_EVENT" );
          break;
        case Event.SWITCH_EVENT:
          sb.append( "SWITCH_EVENT" );
          break;
      }
      
      sb.append( "]" );
      
      return sb.toString();
    }
  }
  
  /*
   * Singleton
   */
  private static Reactor instance;
  
  static {
    instance = null;
  }
  
  public static Reactor i() {
    if ( instance == null )
      instance = new Reactor();
    
    return instance;  
  }
}
